/*
 * This driver is part of lrtbf. It answer the parallel port
 * interrupts by flipping the data twice.
 * Copyright (C) 2005 Kristian Benoit, Opersys inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/fs.h>
#include <linux/spinlock.h>
#include <asm/io.h>


#undef CONFIG_IPIPE_CORE

#ifdef CONFIG_IPIPE_CORE
#warning compiling for IPIPE
#include <linux/ipipe.h>
#define DOMAIN_ENTRY_PARAM void
typedef struct ipipe_domain ipipe_domain_t ;
typedef struct ipipe_domain_attr ipipe_domain_attr_t;
#elif defined(CONFIG_ADEOS)
#warning compiling for ADEOS
#include <linux/adeos.h>

#define CONFIG_IPIPE_CORE 1
#define ipipe_domain_t adomain_t
#define ipipe_domain_attr_t adattr_t
#define ipipe_init_attr adeos_init_attr
#define ipipe_register_domain adeos_register_domain
#define ipipe_unregister_domain adeos_unregister_domain
#define IPIPE_ROOT_PRI ADEOS_ROOT_PRI
#define ipipe_processor_id adeos_processor_id
#define ipipe_virtualize_irq adeos_virtualize_irq
#define ipipe_control_irq adeos_control_irq
#define ipipe_suspend_domain adeos_suspend_domain
#define ipipe_current_domain adp_current
#define DOMAIN_ENTRY_PARAM int iflag

#elif defined(CONFIG_PREEMPT_RT)
#warning compiling for PREEMPT_RT
#else
#warning compiling for VANILLA
#endif

#define PAR_TEST_IRQ 7
#define PAR_TEST_NAME "par_test"
#define PAR_TEST_DEV_ID (void *)par_test_irq_handler
#define PAR_TEST_BASE 0x378

#define PAR_TEST_PORT_DATA PAR_TEST_BASE
#define PAR_TEST_PORT_STATUS PAR_TEST_BASE+1
#define PAR_TEST_PORT_CONTROL PAR_TEST_BASE+2

#define PAR_TEST_CMD_ENABLE_INTR (1<<4)
#define PAR_TEST_CMD_DISABLE_INTR 0

static struct {
	int num_int;
#ifdef CONFIG_IPIPE_CORE
	ipipe_domain_t domain;
#endif //CONFIG_IPIPE_CORE
} par_test;

#ifdef CONFIG_IPIPE_CORE
static void par_test_irq_handler (unsigned);
static void par_test_domain_init (DOMAIN_ENTRY_PARAM);
#else
static irqreturn_t par_test_irq_handler (int, void *, struct pt_regs *);
#endif /*CONFIG_IPIPE_CORE*/

static void __init par_test_do_init (void);

static int __init par_test_init (void)
{
	int retval = 0;
	int Set_region = 0;

#if defined(CONFIG_IPIPE_CORE)
	int Set_domain = 0;
	ipipe_domain_attr_t attr;

	ipipe_init_attr (&attr);
	attr.name		= PAR_TEST_NAME;
	attr.priority	= IPIPE_ROOT_PRI+1;
	attr.entry		= &par_test_domain_init;
#endif /*CONFIG_IPIPE_CORE*/

	if (! request_region (PAR_TEST_BASE, 3, PAR_TEST_NAME)) {
		printk ("%s: Cannot request ioport.\n", PAR_TEST_NAME);
		retval = -1;
		goto clean;
	}
	Set_region = 1;

#ifdef CONFIG_IPIPE_CORE
	if (ipipe_register_domain(&(par_test.domain),&attr)) {
		printk ("%s: Connot register domain (IPIPE).\n", PAR_TEST_NAME);
		retval = -1;
		goto clean;
	}
	Set_domain = 1;
#else
    if (request_irq ( PAR_TEST_IRQ,
					  &par_test_irq_handler,
#if CONFIG_PREEMPT_RT
					  SA_NODELAY,
#else //!CONFIG_PREEMPT_RT
					  SA_INTERRUPT,
#endif //PREEMPT_RT
					  PAR_TEST_NAME,
					  PAR_TEST_DEV_ID )) {
		printk ("%s: Cannot request irq %i\n", PAR_TEST_NAME, PAR_TEST_IRQ);
		retval = -1;
		goto clean;
	} 

#endif /*CONFIG_IPIPE_CORE*/

	par_test_do_init();

clean:
	if (retval){
#ifdef CONFIG_IPIPE_CORE
		if (Set_domain)
			ipipe_unregister_domain (&(par_test.domain));
#endif /*CONFIG_IPIPE_CORE*/
		if (Set_region)
			release_region (PAR_TEST_BASE, 3);
	}
	return retval;
}

static void __init par_test_do_init (void)
{
	printk ("%s: Initializing ...\n", PAR_TEST_NAME);
	par_test.num_int = 0;

	outb (0x00, PAR_TEST_PORT_DATA);

	outb (PAR_TEST_CMD_ENABLE_INTR, PAR_TEST_PORT_CONTROL);
}

#ifdef CONFIG_IPIPE_CORE
static void par_test_domain_init (DOMAIN_ENTRY_PARAM)
{
	printk ( "Domain %s started on CPU#%d.\n",
			 ipipe_current_domain->name,
			 ipipe_processor_id());

#ifdef CONFIG_ADEOS
	if (iflag)
	{
#endif
		ipipe_virtualize_irq ( PAR_TEST_IRQ,
							   &par_test_irq_handler,
							   NULL,
							   IPIPE_ENABLE_MASK | IPIPE_HANDLE_MASK);


		outb (PAR_TEST_CMD_ENABLE_INTR, PAR_TEST_PORT_CONTROL);

#ifdef CONFIG_ADEOS
	}
	while (42)
		ipipe_suspend_domain();
#endif
}

static void par_test_irq_handler (unsigned irq)
{
	outb (0x00, PAR_TEST_PORT_DATA);
	outb (0xFF, PAR_TEST_PORT_DATA);

	par_test.num_int++;

	ipipe_control_irq(irq,0,IPIPE_ENABLE_MASK);
}

#else

static irqreturn_t par_test_irq_handler (int irq, void *dev_id, struct pt_regs *regs)
{
	outb (0x00, PAR_TEST_PORT_DATA);
	outb (0xFF, PAR_TEST_PORT_DATA);

	par_test.num_int++;
	return IRQ_HANDLED;
}
#endif //CONFIG_IPIPE_CORE

static void __exit par_test_exit (void)
{
#ifdef CONFIG_IPIPE_CORE
	ipipe_unregister_domain (&(par_test.domain));
#else
	free_irq (PAR_TEST_IRQ, PAR_TEST_DEV_ID);
#endif //CONFIG_IPIPE_CORE
	printk ("%s: got interrupted %d times.\n", PAR_TEST_NAME, par_test.num_int);
	outb (PAR_TEST_CMD_DISABLE_INTR, PAR_TEST_PORT_CONTROL);
	release_region (PAR_TEST_BASE, 3);
}

module_init(par_test_init);
module_exit(par_test_exit);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Kristian Benoit <kbenoit@opersys.com>");
MODULE_DESCRIPTION("A parallel port driver for testing linux realtime.");
MODULE_SUPPORTED_DEVICE("par-test");

